/*
** ElecDisp.c for XmtdsElecDisp widget in .
** 
** Made by MaxTheDogSays (dubray_f@epita.fr && rancur_v@epita.fr
** Login   <vr@epita.fr>
** 
** Started on  Mon May 16 22:24:08 1994 vr
** Last update Tue Nov  1 03:11:04 1994 hydrix
*/

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <stdio.h>
#include <std/sysdep.h>
#include <std/bit_tab.h>
#include <std/Xstd.h>
#include "parse.h"
#include "ElecDispP.h"

static void			Initialize();
static void			SetValues();
static void			Redisplay();
static void			Destroy();

static XtTimerCallbackProc	Timer();

static void			DrawAction();
static void			AskAction();
static void			GrExpAction();

void				MessageFunc();
void				SprayFunc();
void				PutPixelFunc();
void				StopFunc();
void				ScrollRightFunc();
void				ScrollLeftFunc();
void				ScrollUpFunc();
void				ScrollDownFunc();
void				SetFontFunc();
void				ClearFunc();
void				SpeedFunc();

static cmd_def		cmd_defs[] =
{
{"message",		MessageFunc},
{"spray",		SprayFunc},
{"put-pixel",		PutPixelFunc},
{"stop",		StopFunc},
{"scroll-right",	ScrollRightFunc},
{"scroll-left",		ScrollLeftFunc},
{"scroll-up",		ScrollUpFunc},
{"scroll-down",		ScrollDownFunc},
{"set-font",		SetFontFunc},
{"clear",		ClearFunc},
{"speed",		SpeedFunc}
};

static char	defaultTranslations[] = "<Btn1Down>:	draw()\n\
<Btn2Down>:	ask()\n\
<GrExp>:	grexp()";

static XtActionsRec	actionsList[] =
{
{"draw",	(XtActionProc)DrawAction},
{"ask",		(XtActionProc)AskAction},
{"grexp",	(XtActionProc)GrExpAction}
};

static XtResource	resources[] =
{
{XtNforeground,		XtCForeground,	XtRPixel,	sizeof (Pixel),
    XtOffset(XmtdsElecDispWidget,	xmtds_elec_disp.foreground),
    XtRString,		"yellow"},
{XtNbackground,		XtCBackground,	XtRPixel,	sizeof (Pixel),
    XtOffset(XmtdsElecDispWidget,	xmtds_elec_disp.background),
    XtRString,		"black"},
{XtNshadowColor,	XtCShadowColor,	XtRPixel,	sizeof (Pixel),
    XtOffset(XmtdsElecDispWidget,	xmtds_elec_disp.shadow_color),
    XtRString,		"gray50"},
{XtNprogram,		XtCProgram,	XtRString,	sizeof (char *),
    XtOffset(XmtdsElecDispWidget,	xmtds_elec_disp.program),
    XtRString,		""},
{XtNinterval,		XtCInterval,	XtRInt,		sizeof (int),
    XtOffset(XmtdsElecDispWidget,	xmtds_elec_disp.interval),
    XtRString,		"2000"},
{XtNrasterWidth,	XtCRasterWidth,	XtRInt,		sizeof (int),
    XtOffset(XmtdsElecDispWidget,	xmtds_elec_disp.rast_width),
    XtRString,		"100"},
{XtNrasterHeight,	XtCRasterHeight,XtRInt,		sizeof (int),
    XtOffset(XmtdsElecDispWidget,	xmtds_elec_disp.rast_height),
    XtRString,		"16"},
{XtNpixWidth,		XtCPixWidth,	XtRInt,		sizeof (int),
    XtOffset(XmtdsElecDispWidget,	xmtds_elec_disp.pix_width),
    XtRString,		"2"},
{XtNpixHeight,		XtCPixHeight,XtRInt,		sizeof (int),
    XtOffset(XmtdsElecDispWidget,	xmtds_elec_disp.pix_height),
    XtRString,		"2"},
};

XmtdsElecDispClassRec	xmtdsElecDispClassRec = {
{
   /* core */
   (WidgetClass)&widgetClassRec,	/* superclass		*/
   "XmtdsElecDisp",			/* class name		*/
   sizeof(XmtdsElecDispRec),		/* size			*/   
   NULL,				/* class initialize	*/   
   NULL,				/* class_part_init	*/   
   FALSE,				/* class_inited		*/   
   Initialize,				/* initialize		*/   
   NULL,				/* initialize_hook	*/   
   XtInheritRealize,			/* realize		*/   
   actionsList,				/* actions		*/   
   XtNumber(actionsList),		/* num_actions		*/   
   resources,				/* resources		*/   
   XtNumber(resources),			/* resource_count	*/   
   NULLQUARK,				/* xrm_class		*/   
   TRUE,				/* compress_motion	*/   
   TRUE,				/* compress_exposure	*/   
   TRUE,				/* compress_enterleave	*/   
   FALSE,				/* visible_interest	*/   
   Destroy,				/* destroy		*/   
   NULL,				/* resize		*/   
   Redisplay,				/* expose		*/   
   NULL,				/* set_values		*/   
   NULL,				/* set_values_hook	*/   
   XtInheritSetValuesAlmost,		/* set_values_almost	*/   
   NULL,				/* get_values_hook	*/   
   NULL,				/* accept_focus		*/   
   XtVersion,				/* version		*/   
   NULL,				/* callback_private	*/   
   defaultTranslations,			/* tm_table		*/   
   XtInheritQueryGeometry,		/* query_geometry	*/   
   XtInheritDisplayAccelerator,		/* display_accelerator	*/   
   NULL,				/* extension		*/   
},
/* xmtds_elec_disp */
{
   0,					/* empty		*/
}
};

WidgetClass	xmtdsElecDispWidgetClass = (WidgetClass)&xmtdsElecDispClassRec;

static void		Initialize(request,new,args,num_args)
XmtdsElecDispWidget	request;
XmtdsElecDispWidget	new;
ArgList			args;
Cardinal		*num_args;
{
   XGCValues		xgcv;
   
   xgcv.foreground = new->xmtds_elec_disp.foreground;
   new->xmtds_elec_disp.on_gc =
      XtGetGC(new,
	      GCForeground,
	      &xgcv);
   xgcv.foreground = new->xmtds_elec_disp.shadow_color;
   new->xmtds_elec_disp.off_gc =
      XtGetGC(new,
	      GCForeground,
	      &xgcv);
   xgcv.function = GXcopy;
   xgcv.foreground = W_BPX(new);
   new->xmtds_elec_disp.black_gc =
      XtGetGC(new,
	      GCForeground,
	      &xgcv);
   xgcv.function = GXcopy;
   new->xmtds_elec_disp.copy_gc =
      XtGetGC(new,
	      GCFunction,
	      &xgcv);
   new->xmtds_elec_disp.bgp =
      XCreatePixmap(XtDisplay(new),
		    W_ROOT(new),
		    new->xmtds_elec_disp.pix_width*2,
		    new->xmtds_elec_disp.pix_height*2,
		    W_DEPTH(new));
   XFillRectangle(XtDisplay(new),
		  new->xmtds_elec_disp.bgp,
		  new->xmtds_elec_disp.black_gc,
		  0,
		  0,
		  new->xmtds_elec_disp.pix_width*2,
		  new->xmtds_elec_disp.pix_height*2);
   XFillRectangle(XtDisplay(new),
		  new->xmtds_elec_disp.bgp,
		  new->xmtds_elec_disp.off_gc,
		  new->xmtds_elec_disp.pix_width,
		  new->xmtds_elec_disp.pix_height,
		  new->xmtds_elec_disp.pix_width,
		  new->xmtds_elec_disp.pix_height);
   new->core.background_pixmap = new->xmtds_elec_disp.bgp;
   new->xmtds_elec_disp.xt_interval_id =
      XtAppAddTimeOut(XtWidgetToApplicationContext((Widget)new),
		      new->xmtds_elec_disp.interval,
		      Timer,
		      new);
   bit_tab_create(&(new->xmtds_elec_disp.video_mem),
		  new->xmtds_elec_disp.rast_width,
		  new->xmtds_elec_disp.rast_height,
		  &(new->xmtds_elec_disp.video_mem_size));
   bzero(new->xmtds_elec_disp.video_mem,
	 new->xmtds_elec_disp.video_mem_size);
   new->core.width =
      new->xmtds_elec_disp.rast_width*(new->xmtds_elec_disp.pix_width*2);
   new->core.height =
      new->xmtds_elec_disp.rast_height*(new->xmtds_elec_disp.pix_height*2);
   parse_compile(&(new->xmtds_elec_disp.ps),
		 cmd_defs,
		 XtNumber(cmd_defs),
		 new->xmtds_elec_disp.program,
		 strlen(new->xmtds_elec_disp.program),
		 new);
   parse_rewind(&(new->xmtds_elec_disp.ps));
   if ((new->xmtds_elec_disp.message_font = 
	XLoadQueryFont(XtDisplay(new),DEFAULT_MESSAGE_FONT)) == NULL)
   {
      XtWarning("Unknown font");
      exit(-1);
   }
   new->xmtds_elec_disp.stop = FALSE;
}

static void		SetValues(old,request,new,args,num_args)
XmtdsElecDispWidget	old;
XmtdsElecDispWidget	request;
XmtdsElecDispWidget	new;
ArgList			args;
Cardinal		*num_args;
{
   if (request->xmtds_elec_disp.program != old->xmtds_elec_disp.program)
   {
      parse_destroy(&(new->xmtds_elec_disp.ps)); 
      free(old->xmtds_elec_disp.program);
      parse_compile(&(new->xmtds_elec_disp.ps),
		    cmd_defs,
		    XtNumber(cmd_defs),
		    new->xmtds_elec_disp.program,
		    strlen(new->xmtds_elec_disp.program),
		    new);
      parse_rewind(&(new->xmtds_elec_disp.ps));  
   }
   if ((request->xmtds_elec_disp.rast_width !=
	old->xmtds_elec_disp.rast_width) ||
       (request->xmtds_elec_disp.rast_height !=
	old->xmtds_elec_disp.rast_height))
   {
      free(old->xmtds_elec_disp.video_mem);
      bit_tab_create(&(new->xmtds_elec_disp.video_mem),
		     new->xmtds_elec_disp.rast_width,
		     new->xmtds_elec_disp.rast_height,
		     &(new->xmtds_elec_disp.video_mem_size));
      bzero(new->xmtds_elec_disp.video_mem,
	    new->xmtds_elec_disp.video_mem_size);
   }
}

void			refresh_raster(w,x,y,width,height)
XmtdsElecDispWidget	w;
int			x;
int			y;
int			width;
int			height;
{
   XRectangle		on_xrects[XRECT_BUFFER];
   XRectangle		off_xrects[XRECT_BUFFER];
   int			num_on_xrects;
   int			num_off_xrects;
   int			i;
   int			dest_i;
   int			j;
   int			dest_j;
   
   num_on_xrects = 0;
   num_off_xrects = 0;
   j = y;
   dest_i = x+width;
   dest_j = y+height;
   if (dest_i > w->xmtds_elec_disp.rast_width)
      dest_i = w->xmtds_elec_disp.rast_width;
   if (dest_j > w->xmtds_elec_disp.rast_height)
      dest_j = w->xmtds_elec_disp.rast_height;
   while (j < dest_j)
   {
      i = x;
      while (i < dest_i)
      {
	 if (BIT_TEST(w->xmtds_elec_disp.video_mem,
		      i,
		      j,
		      w->xmtds_elec_disp.rast_width))
	 {
	    if (num_on_xrects == XRECT_BUFFER)
	    {
	       XFillRectangles(XtDisplay(w),
			       XtWindow(w),
			       w->xmtds_elec_disp.on_gc,
			       on_xrects,
			       num_on_xrects);
	       num_on_xrects = 0;
	    }
	    on_xrects[num_on_xrects].x =
	       i*(w->xmtds_elec_disp.pix_width*2)+w->xmtds_elec_disp.pix_width;
	    on_xrects[num_on_xrects].y =
	       j*(w->xmtds_elec_disp.pix_height*2)+w->xmtds_elec_disp.pix_height;
	    on_xrects[num_on_xrects].width = w->xmtds_elec_disp.pix_width;
	    on_xrects[num_on_xrects].height = w->xmtds_elec_disp.pix_height;
	    num_on_xrects++;
	 }
	 else
	 {
	    if (num_off_xrects == XRECT_BUFFER)
	    {
	       XFillRectangles(XtDisplay(w),
			       XtWindow(w),
			       w->xmtds_elec_disp.off_gc,
			       off_xrects,
			       num_off_xrects);
	       num_off_xrects = 0;
	    }
	    off_xrects[num_off_xrects].x =
	       i*(w->xmtds_elec_disp.pix_width*2)+w->xmtds_elec_disp.pix_width;
	    off_xrects[num_off_xrects].y =
	       j*(w->xmtds_elec_disp.pix_height*2)+w->xmtds_elec_disp.pix_height;
	    off_xrects[num_off_xrects].width = w->xmtds_elec_disp.pix_width;
	    off_xrects[num_off_xrects].height = w->xmtds_elec_disp.pix_height;
	    num_off_xrects++;
	 }
	 i++;
      }
      j++;
   }
   XFillRectangles(XtDisplay(w),
		   XtWindow(w),
		   w->xmtds_elec_disp.on_gc,
		   on_xrects,
		   num_on_xrects);
   XFillRectangles(XtDisplay(w),
		   XtWindow(w),
		   w->xmtds_elec_disp.off_gc,
		   off_xrects,
		   num_off_xrects);
}

static void		Redisplay(w,event,region)
XmtdsElecDispWidget	w;
XExposeEvent		*event;
Region			region;
{
   refresh_raster(w,
		  event->x/(w->xmtds_elec_disp.pix_width*2),
		  event->y/(w->xmtds_elec_disp.pix_height*2),
		  event->width/(w->xmtds_elec_disp.pix_width*2)+2,
		  event->height/(w->xmtds_elec_disp.pix_height*2)+2);
}

static void		Destroy(w)
XmtdsElecDispWidget	w;
{
   XtReleaseGC(w,w->xmtds_elec_disp.on_gc);
   XtReleaseGC(w,w->xmtds_elec_disp.off_gc);
   XtReleaseGC(w,w->xmtds_elec_disp.copy_gc);
   XtReleaseGC(w,w->xmtds_elec_disp.black_gc);
   XtRemoveTimeOut(w->xmtds_elec_disp.xt_interval_id);
   XFreePixmap(XtDisplay(w),w->xmtds_elec_disp.bgp);
   XFreeFont(XtDisplay(w),w->xmtds_elec_disp.message_font);
   free(w->xmtds_elec_disp.video_mem);
   parse_destroy(&(w->xmtds_elec_disp.ps));  
}

static XtTimerCallbackProc	Timer(w,id)
XmtdsElecDispWidget		w;
XtIntervalId			*id;
{
   if (w->xmtds_elec_disp.stop)
   {
      XtRemoveTimeOut(w->xmtds_elec_disp.xt_interval_id);
      return ;
   }
   if (parse_next(&(w->xmtds_elec_disp.ps)) == PARSE_EOP)
   {
      parse_rewind(&(w->xmtds_elec_disp.ps));
      parse_next(&(w->xmtds_elec_disp.ps));
   }
   w->xmtds_elec_disp.xt_interval_id =
      XtAppAddTimeOut(XtWidgetToApplicationContext((Widget)w),
		      w->xmtds_elec_disp.interval,
		      Timer,
		      w);  
}

static void		DrawAction(w,event,args,n_args)
XmtdsElecDispWidget	w;
XButtonEvent		*event;
String			*args;
int			*n_args;
{
   int			x;
   int			y;
   
   x = event->x/(w->xmtds_elec_disp.pix_width*2);
   y = event->y/(w->xmtds_elec_disp.pix_height*2);
   if ((x > w->xmtds_elec_disp.rast_width) ||
       (y > w->xmtds_elec_disp.rast_height))
      return ;
   if (event->type != ButtonPress)
      return ;
   BIT_SET(w->xmtds_elec_disp.video_mem,
	   x,
	   y,
	   w->xmtds_elec_disp.rast_width);
   XFillRectangle(XtDisplay(w),
		  XtWindow(w),
		  w->xmtds_elec_disp.on_gc,
		  x*(w->xmtds_elec_disp.pix_width*2)+
		  w->xmtds_elec_disp.pix_width,
		  y*(w->xmtds_elec_disp.pix_height*2)+
		  w->xmtds_elec_disp.pix_height,
		  w->xmtds_elec_disp.pix_width,
		  w->xmtds_elec_disp.pix_height);
}

static void		AskAction(w,event,args,n_args)
XmtdsElecDispWidget	w;
XButtonEvent		*event;
String			*args;
int			*n_args;
{
   int			x;
   int			y;
   
   x = event->x/(w->xmtds_elec_disp.pix_width*2);
   y = event->y/(w->xmtds_elec_disp.pix_height*2);
   if ((x > w->xmtds_elec_disp.rast_width) ||
       (y > w->xmtds_elec_disp.rast_height))
      return ;
   if (event->type != ButtonPress)
      return ;
   printf("x = %d y = %d offset: %d mask: %d bit = %d\n",
	  x,
	  y,
	  BIT_OFFSET(x,
		     y,
		     w->xmtds_elec_disp.rast_width),
	  BIT_MASK(x),
	  BIT_TEST(w->xmtds_elec_disp.video_mem,
		   x,
		   y,
		   w->xmtds_elec_disp.rast_width));
}

static void		GrExpAction(w,event,args,n_args)
XmtdsElecDispWidget	w;
XGraphicsExposeEvent	*event;
String			*args;
int			*n_args;
{
   if (event->type != GraphicsExpose)
      return ;
   refresh_raster(w,
		  event->x/(w->xmtds_elec_disp.pix_width*2),
		  event->y/(w->xmtds_elec_disp.pix_height*2),
		  event->width/(w->xmtds_elec_disp.pix_width*2)+2,
		  event->height/(w->xmtds_elec_disp.pix_height*2)+2);
}

void			MessageFunc(w,args,num_args)
XmtdsElecDispWidget	w;
char			**args;
int			num_args;
{
   int			x;
   int			y;
   int			width;
   int			height;
   int			len;
   GC			gc;
   XGCValues		xgcv;
   XImage		*image;
   int			i;
   int			j;
   XRectangle		on_xrects[XRECT_BUFFER];
   XRectangle		off_xrects[XRECT_BUFFER];
   int			num_on_xrects;
   int			num_off_xrects;
   Pixmap		pix;
   
   num_on_xrects = 0;
   num_off_xrects = 0;
   if (num_args != 3)
   {
      XtWarning("syntax: message(x,y,str)");
      exit(-1);
   }
   x = atoi(args[0]);
   y = atoi(args[1]);
   len = strlen(args[2]);
   width = XTextWidth(w->xmtds_elec_disp.message_font,args[2],len);
   height = XFS_HEIGHT(w->xmtds_elec_disp.message_font);
   if ((x+width) > w->xmtds_elec_disp.rast_width)
      width = w->xmtds_elec_disp.rast_width-x;
   if ((y+height) > w->xmtds_elec_disp.rast_height)
      height = w->xmtds_elec_disp.rast_height-y;
   pix = XCreatePixmap(XtDisplay(w),
		       W_ROOT(w),
		       width,
		       height,
		       1);
   xgcv.foreground = W_WPX(w);
   xgcv.font = XFS_FID(w->xmtds_elec_disp.message_font);
   gc = XCreateGC(XtDisplay(w),
		  pix,
		  GCForeground|GCFont,
		  &xgcv);
   XFillRectangle(XtDisplay(w),
		  pix,
		  gc,
		  0,
		  0,
		  width,
		  height);
   XSetForeground(XtDisplay(w),
		  gc,
		  W_BPX(w));
   XDrawString(XtDisplay(w),
	       pix,
	       gc,
	       0,
	       XFS_HEIGHT(w->xmtds_elec_disp.message_font)-
	       XFS_DESCENT(w->xmtds_elec_disp.message_font),
	       args[2],
	       len);
   image = XGetImage(XtDisplay(w),
		     pix,
		     0,
		     0,
		     width,
		     height,
		     -1,
		     ZPixmap);
   j = 0;
   while (j < height)
   {
      i = 0;
      while (i < width)
      {
	 if (XGetPixel(image,i,j) == W_BPX(w))
	 {
	    BIT_SET(w->xmtds_elec_disp.video_mem,
		    x+i,
		    y+j,
		    w->xmtds_elec_disp.rast_width);
	    if (num_on_xrects == XRECT_BUFFER)
	    {
	       XFillRectangles(XtDisplay(w),
			       XtWindow(w),
			       w->xmtds_elec_disp.on_gc,
			       on_xrects,
			       num_on_xrects);
	       num_on_xrects = 0;
	    }
	    on_xrects[num_on_xrects].x =
	       (x+i)*(w->xmtds_elec_disp.pix_width*2)+
		  w->xmtds_elec_disp.pix_width;
	    on_xrects[num_on_xrects].y =
	       (y+j)*(w->xmtds_elec_disp.pix_height*2)+
		  w->xmtds_elec_disp.pix_height;
	    on_xrects[num_on_xrects].width = w->xmtds_elec_disp.pix_width;
	    on_xrects[num_on_xrects].height = w->xmtds_elec_disp.pix_height;
	    num_on_xrects++;
	 }
	 else
	 {
	    BIT_UNSET(w->xmtds_elec_disp.video_mem,
		      x+i,
		      y+j,
		      w->xmtds_elec_disp.rast_width);
	    if (num_off_xrects == XRECT_BUFFER)
	    {
	       XFillRectangles(XtDisplay(w),
			       XtWindow(w),
			       w->xmtds_elec_disp.off_gc,
			       off_xrects,
			       num_off_xrects);
	       num_off_xrects = 0;
	    }
	    off_xrects[num_off_xrects].x =
	       (x+i)*(w->xmtds_elec_disp.pix_width*2)+
		  w->xmtds_elec_disp.pix_width;
	    off_xrects[num_off_xrects].y =
	       (y+j)*(w->xmtds_elec_disp.pix_height*2)+
		  w->xmtds_elec_disp.pix_height;
	    off_xrects[num_off_xrects].width = w->xmtds_elec_disp.pix_width;
	    off_xrects[num_off_xrects].height = w->xmtds_elec_disp.pix_height;
	    num_off_xrects++;
	 }
	 i++;
      }
      j++;
   }
   XFillRectangles(XtDisplay(w),
		   XtWindow(w),
		   w->xmtds_elec_disp.on_gc,
		   on_xrects,
		   num_on_xrects);
   XFillRectangles(XtDisplay(w),
		   XtWindow(w),
		   w->xmtds_elec_disp.off_gc,
		   off_xrects,
		   num_off_xrects);
   XFreeGC(XtDisplay(w),gc);
   XFreePixmap(XtDisplay(w),pix);
   XDestroyImage(image);
}

static int		myrand(n)
int			n;
{
   return (rand()%n);
}

static void		spray(w,xr,num_xr)
XmtdsElecDispWidget	w;
XRectangle		*xr;
int			num_xr;
{
   Boolean		spray[XRECT_BUFFER];
   int			i;
   int			j;
   int			ran;
   int			found;
   
   i = 0;
   while (i < XRECT_BUFFER)
      spray[i++] = 0;
   i = 0;
   while (i < num_xr)
   {
      ran = myrand(XRECT_BUFFER);
      found = 0;
      j = ran;
      while (j < num_xr)
      {
	 if (!spray[j])
	 {
	    ran = j;
	    spray[ran] = 1;
	    XFillRectangle(XtDisplay(w),
			   XtWindow(w),
			   w->xmtds_elec_disp.on_gc,
			   xr[ran].x,
			   xr[ran].y,
			   xr[ran].width,
			   xr[ran].height);
	    XFlush(XtDisplay(w));
	    found = 1;
	    break ;
	 }
	 j++;
      }
      if (found)
	 i++;
   }
}

void			SprayFunc(w,args,num_args)
XmtdsElecDispWidget	w;
char			**args;
int			num_args;
{
   int			x;
   int			y;
   int			width;
   int			height;
   int			len;
   GC			gc;
   XGCValues		xgcv;
   XImage		*image;
   int			i;
   int			j;
   XRectangle		on_xrects[XRECT_BUFFER];
   XRectangle		off_xrects[XRECT_BUFFER];
   int			num_on_xrects;
   int			num_off_xrects;
   Pixmap		pix;
   
   num_on_xrects = 0;
   num_off_xrects = 0;
   if (num_args != 3)
   {
      XtWarning("syntax: message(x,y,str)");
      exit(-1);
   }
   x = atoi(args[0]);
   y = atoi(args[1]);
   len = strlen(args[2]);
   width = XTextWidth(w->xmtds_elec_disp.message_font,args[2],len);
   height = XFS_HEIGHT(w->xmtds_elec_disp.message_font);
   if ((x+width) > w->xmtds_elec_disp.rast_width)
      width = w->xmtds_elec_disp.rast_width-x;
   if ((y+height) > w->xmtds_elec_disp.rast_height)
      height = w->xmtds_elec_disp.rast_height-y;
   pix = XCreatePixmap(XtDisplay(w),
		       W_ROOT(w),
		       width,
		       height,
		       1);
   xgcv.foreground = W_WPX(w);
   xgcv.font = XFS_FID(w->xmtds_elec_disp.message_font);
   gc = XCreateGC(XtDisplay(w),
		  pix,
		  GCForeground|GCFont,
		  &xgcv);
   XFillRectangle(XtDisplay(w),
		  pix,
		  gc,
		  0,
		  0,
		  width,
		  height);
   XSetForeground(XtDisplay(w),
		  gc,
		  W_BPX(w));
   XDrawString(XtDisplay(w),
	       pix,
	       gc,
	       0,
	       XFS_HEIGHT(w->xmtds_elec_disp.message_font)-
	       XFS_DESCENT(w->xmtds_elec_disp.message_font),
	       args[2],
	       len);
   image = XGetImage(XtDisplay(w),
		     pix,
		     0,
		     0,
		     width,
		     height,
		     -1,
		     ZPixmap);
   j = 0;
   while (j < height)
   {
      i = 0;
      while (i < width)
      {
	 if (XGetPixel(image,i,j) == W_BPX(w))
	 {
	    BIT_SET(w->xmtds_elec_disp.video_mem,
		    x+i,
		    y+j,
		    w->xmtds_elec_disp.rast_width);
	    if (num_on_xrects == XRECT_BUFFER)
	    {
	       spray(w,on_xrects,num_on_xrects);
	       num_on_xrects = 0;
	    }
	    on_xrects[num_on_xrects].x =
	       (x+i)*(w->xmtds_elec_disp.pix_width*2)+
		  w->xmtds_elec_disp.pix_width;
	    on_xrects[num_on_xrects].y =
	       (y+j)*(w->xmtds_elec_disp.pix_height*2)+
		  w->xmtds_elec_disp.pix_height;
	    on_xrects[num_on_xrects].width = w->xmtds_elec_disp.pix_width;
	    on_xrects[num_on_xrects].height = w->xmtds_elec_disp.pix_height;
	    num_on_xrects++;
	 }
	 else
	 {
	    BIT_UNSET(w->xmtds_elec_disp.video_mem,
		      x+i,
		      y+j,
		      w->xmtds_elec_disp.rast_width);
	    if (num_off_xrects == XRECT_BUFFER)
	    {
	       XFillRectangles(XtDisplay(w),
			       XtWindow(w),
			       w->xmtds_elec_disp.off_gc,
			       off_xrects,
			       num_off_xrects);
	       num_off_xrects = 0;
	    }
	    off_xrects[num_off_xrects].x =
	       (x+i)*(w->xmtds_elec_disp.pix_width*2)+
		  w->xmtds_elec_disp.pix_width;
	    off_xrects[num_off_xrects].y =
	       (y+j)*(w->xmtds_elec_disp.pix_height*2)+
		  w->xmtds_elec_disp.pix_height;
	    off_xrects[num_off_xrects].width = w->xmtds_elec_disp.pix_width;
	    off_xrects[num_off_xrects].height = w->xmtds_elec_disp.pix_height;
	    num_off_xrects++;
	 }
	 i++;
      }
      j++;
   }
   spray(w,on_xrects,num_on_xrects);
   XFillRectangles(XtDisplay(w),
		   XtWindow(w),
		   w->xmtds_elec_disp.off_gc,
		   off_xrects,
		   num_off_xrects);
   XFreeGC(XtDisplay(w),gc);
   XFreePixmap(XtDisplay(w),pix);
   XDestroyImage(image);
}

void			PutPixelFunc(w,args,num_args)
XmtdsElecDispWidget	w;
char			**args;
int			num_args;
{
   int			x;
   int			y;
   int			lit;
   GC			gc;
   
   if (num_args != 3)
   {
      XtWarning("syntax: put-pixel(x,y,0/1)");
      exit(-1);
   }
   x = atoi(args[0]);
   y = atoi(args[1]);
   lit = atoi(args[2]);
   if (lit)
   {
      BIT_SET(w->xmtds_elec_disp.video_mem,
	      x,
	      y,
	      w->xmtds_elec_disp.rast_width);
      gc = w->xmtds_elec_disp.on_gc;
   }
   else
   {
      BIT_UNSET(w->xmtds_elec_disp.video_mem,
		x,
		y,
		w->xmtds_elec_disp.rast_width);
      gc = w->xmtds_elec_disp.off_gc;  
   }
   XFillRectangle(XtDisplay(w),
		  XtWindow(w),
		  gc,
		  x*(w->xmtds_elec_disp.pix_width*2)+
		  w->xmtds_elec_disp.pix_width,
		  y*(w->xmtds_elec_disp.pix_height*2)+
		  w->xmtds_elec_disp.pix_height,
		  w->xmtds_elec_disp.pix_width,
		  w->xmtds_elec_disp.pix_height);
}

void			StopFunc(w,args,num_args)
XmtdsElecDispWidget	w;
char			**args;
int			num_args;
{
   w->xmtds_elec_disp.stop = TRUE;
}

void			ScrollRightFunc(w,args,num_args)
XmtdsElecDispWidget	w;
char			**args;
int			num_args;
{
   XCopyArea(XtDisplay(w),
	     XtWindow(w),
	     XtWindow(w),
	     w->xmtds_elec_disp.copy_gc,
	     0,
	     0,
	     w->xmtds_elec_disp.rast_width*w->xmtds_elec_disp.pix_width*2,
	     w->xmtds_elec_disp.rast_height*w->xmtds_elec_disp.pix_height*2,
	     (w->xmtds_elec_disp.pix_width*2),
	     0);
   bit_tab_scroll_right(w->xmtds_elec_disp.video_mem,
			w->xmtds_elec_disp.rast_width,
			w->xmtds_elec_disp.rast_height);
}

void			ScrollLeftFunc(w,args,num_args)
XmtdsElecDispWidget	w;
char			**args;
int			num_args;
{
   XCopyArea(XtDisplay(w),
	     XtWindow(w),
	     XtWindow(w),
	     w->xmtds_elec_disp.copy_gc,
	     (w->xmtds_elec_disp.pix_width*2),
	     0,
	     w->xmtds_elec_disp.rast_width*w->xmtds_elec_disp.pix_width*2,
	     w->xmtds_elec_disp.rast_height*w->xmtds_elec_disp.pix_height*2,
	     0,
	     0);
   bit_tab_scroll_left(w->xmtds_elec_disp.video_mem,
		       w->xmtds_elec_disp.rast_width,
		       w->xmtds_elec_disp.rast_height);
}

void			ScrollUpFunc(w,args,num_args)
XmtdsElecDispWidget	w;
char			**args;
int			num_args;
{
   XCopyArea(XtDisplay(w),
	     XtWindow(w),
	     XtWindow(w),
	     w->xmtds_elec_disp.copy_gc,
	     0,
	     (w->xmtds_elec_disp.pix_height*2),
	     w->xmtds_elec_disp.rast_width*w->xmtds_elec_disp.pix_width*2,
	     w->xmtds_elec_disp.rast_height*w->xmtds_elec_disp.pix_height*2,
	     0,
	     0);
   bit_tab_scroll_up(w->xmtds_elec_disp.video_mem,
		     w->xmtds_elec_disp.rast_width,
		     w->xmtds_elec_disp.rast_height);
}

void			ScrollDownFunc(w,args,num_args)
XmtdsElecDispWidget	w;
char			**args;
int			num_args;
{
   XCopyArea(XtDisplay(w),
	     XtWindow(w),
	     XtWindow(w),
	     w->xmtds_elec_disp.copy_gc,
	     0,
	     0,
	     w->xmtds_elec_disp.rast_width*w->xmtds_elec_disp.pix_width*2,
	     w->xmtds_elec_disp.rast_height*w->xmtds_elec_disp.pix_height*2,
	     0,
	     w->xmtds_elec_disp.pix_height*2);
   bit_tab_scroll_down(w->xmtds_elec_disp.video_mem,
		       w->xmtds_elec_disp.rast_width,
		       w->xmtds_elec_disp.rast_height);
}

void			SetFontFunc(w,args,num_args)
XmtdsElecDispWidget	w;
char			**args;
int			num_args;
{
   if (num_args != 1)
   {
      XtWarning("syntax: set-font(font name)");
      exit(-1);
   }
   XFreeFont(XtDisplay(w),
	     w->xmtds_elec_disp.message_font);
   if ((w->xmtds_elec_disp.message_font = 
	XLoadQueryFont(XtDisplay(w),args[0])) == NULL)
   {
      XtWarning("Unknown font");
      exit(-1);
   }  
}

void			ClearFunc(w,args,num_args)
XmtdsElecDispWidget	w;
char			**args;
int			num_args;
{
   XRectangle		off_xrects[XRECT_BUFFER];
   int			num_off_xrects;
   int			i;
   int			j;
   
   num_off_xrects = 0;
   j = 0;
   while (j < w->xmtds_elec_disp.rast_height)
   {
      i = 0;
      while (i < w->xmtds_elec_disp.rast_width)
      {
	 if (BIT_TEST(w->xmtds_elec_disp.video_mem,
		      i,
		      j,
		      w->xmtds_elec_disp.rast_width))
	 {
	    if (num_off_xrects == XRECT_BUFFER)
	    {
	       XFillRectangles(XtDisplay(w),
			       XtWindow(w),
			       w->xmtds_elec_disp.off_gc,
			       off_xrects,
			       num_off_xrects);
	       num_off_xrects = 0;
	    }
	    off_xrects[num_off_xrects].x =
	       i*(w->xmtds_elec_disp.pix_width*2)+w->xmtds_elec_disp.pix_width;
	    off_xrects[num_off_xrects].y =
	       j*(w->xmtds_elec_disp.pix_height*2)+w->xmtds_elec_disp.pix_height;
	    off_xrects[num_off_xrects].width = w->xmtds_elec_disp.pix_width;
	    off_xrects[num_off_xrects].height = w->xmtds_elec_disp.pix_height;
	    num_off_xrects++;
	 }
	 i++;
      }
      j++;
   }
   XFillRectangles(XtDisplay(w),
		   XtWindow(w),
		   w->xmtds_elec_disp.off_gc,
		   off_xrects,
		   num_off_xrects);
   bzero(w->xmtds_elec_disp.video_mem,
	 w->xmtds_elec_disp.video_mem_size);  
}

void			SpeedFunc(w,args,num_args)
XmtdsElecDispWidget	w;
char			**args;
int			num_args;
{
   int			time;

   if (num_args != 1)
   {
      XtWarning("syntax: speed(time)");
      exit(-1);
   }
   time = atoi(args[0]);
   w->xmtds_elec_disp.interval = time;  
}
